%% This code accompanies the manuscript
%% Riccardo Bardin, Fleurianne Bertrand, Olena Palii, Matthias Schlottbom
%% On a phase-space discontinuous Galerkin approximation for the radiative transfer equation in slab geometry.
%% Computational Methods in Applied Mathematics, doi: 10.1515/cmam-2023-0090
%% https://arxiv.org/abs/2201.06104

% Solve RTE in slab geometry for subsquently refined meshes.
plotting = 0;                       % plotting: 0 -> off, 1 -> on
uniform = 1;                        % mesh refinement: 0 -> adaptive, 1 -> uniform
estimator = 'hierarchical_h';       % 'hierarchical_h' or 'hierarchical_p'
test_case = 1;                      % Possible choices: 1 (-> Section 4), 2 (-> Section 5, (26)), 3 (-> Section 5, (27))
output = 0;                         % output internal message in the solver.
Theta = 0.3;                        % Doerfler marking parameter. Theta*100 percent of the error will be in the refined elements.

k = 1; % specify polynomial degrees (here k_mu=k and k_z=k+1)

%% Initialize error arrays
GalL2 = [];
GalVh = [];
dofs = [];

estimator_values = [];
estimator_valuesL2 = [];
estimator_valuesH1 = [];

%% Local polynomial degree no_basis-1, implemented up to 6 (max admissible k = 4)
no_basis_z = k+2;  % z-coordinate, from 1 to 6 = degree+1
no_basis_mu = k+1; % mu-coordinate, from 1 to 6 = degree+1

%% Extra bsis functions for hierarchical error estimators
if strcmp(estimator,'hierarchical_p')
    extra_mu = 1;
    extra_z = 1;
else
    extra_mu = 0;
    extra_z = 0;
end

%% Load Test Cases
if test_case == 1
    [ux, dux, d2ux, qx, gLx, gRx, sigmat, sigmas] = loadTestCase1();       % discontinuity at mu=1/2, but 'smooth'.
elseif test_case == 2
    [ux, dux, d2ux, qx, gLx, gRx, sigmat, sigmas] = loadTestCase2();       % |x+mu|^{1/2} % all derivatives exist, but become unbounded or non-integrable
elseif test_case == 3
    [ux, dux, d2ux, qx, gLx, gRx, sigmat, sigmas] = loadTestCase3();       % discontinuity at mu=1/sqrt(2).
end

%% Initialize quadtree-grid
elem = [1; 2; 3; 4; 0; 1];
edges = [1, 2; 4, 3];
nodes = [0, 1, 1, 0; 0, 0, 1, 1];
[edge2elem, int_edges] = edge_to_elem(elem, nodes, edges);

%% Plot mesh
% figure(1)
% patch('Faces', elem(1:4, :)', 'Vertices', nodes', 'FaceColor', 'interp', 'FaceVertexCData', ones(size(nodes, 2), 1), 'EdgeAlpha', 0.5, 'LineWidth', 1.25)
% pause

refine = 1; % elements to be refined. Here, refine initial mesh once.

%% Run solver for different meshes
while size(elem, 2) < 4096

    %% refine grid in tbr elements
    for tbr = refine(:)'
        [elem, nodes, edges, edge2elem, int_edges] = refine_quad(elem, nodes, edges, edge2elem, tbr);
    end

    %% Display grid
    if plotting
        figure(1)
        clf()
        patch('Faces', elem(1:4, :)', 'Vertices', nodes', 'FaceColor', 'flat', 'FaceVertexCData', elem(5, :)', 'EdgeAlpha', 0.5, 'LineWidth', 1.25)
        title('Local level of mesh', 'interpreter', 'Latex')
        colorbar
        title_picture = 'Local level of mesh.png';
        print(title_picture, '-dpng','-r300')
    end

    no_basis = no_basis_z*no_basis_mu; % total number of basis functions per element
    [phi, dphi] = load_basis_functions();
    [x,w] = lgwt(12, 0, 1); % Gauss-Legendre integration rule on (0,1) used of spatial integration.

    if uniform == 1
        Cie = get_Cie(no_basis_z-1, phi, dphi, x, w); % compute constant from inverse inequality, cf (17)
        Cdt = 1+sqrt(Cie); % compute discrete trace constant, cf Lemma 2.
        alpha_penalty = 1/2+Cdt;
        % solve variational problem:
        [u, x, w, xc, wc, phi, dphi, Jmp,IP, Penalty, int_edges, edge2elem] = solve_SIP(elem, nodes, edges, sigmas, sigmat, qx, gLx, gRx, no_basis_z, no_basis_mu, alpha_penalty, output, edge2elem, int_edges);

    elseif uniform == 0 && strcmp(estimator, 'hierarchical_p')
        Cie = get_Cie(no_basis_z-1+extra_z, phi, dphi, x, w); % compute constant from inverse inequality, cf (17)
        Cdt = 1+sqrt(Cie); % compute discrete trace constant, cf Lemma 2.
        alpha_penalty = 1/2+Cdt;
        % solve variational problem: for chosen polynomial degree
        [u, x, w, xc, wc, phi, dphi, Jmp, IP, Penalty, int_edges, edge2elem, HF, sys] = solve_SIP(elem, nodes, edges, sigmas, sigmat, qx, gLx, gRx, no_basis_z,         no_basis_mu,          alpha_penalty, output, edge2elem, int_edges);
        % solve variational problem: for increased polynomial degree to
        % compute error estimator
        [uest, ~, ~, ~, ~, ~, ~, Jmp_fine, IPf, Penalty_fine, ~, ~, ~, sys_fine]           = solve_SIP(elem, nodes, edges, sigmas, sigmat, qx, gLx, gRx, no_basis_z+extra_z, no_basis_mu+extra_mu, alpha_penalty, output, edge2elem, int_edges);
        % compute error estimator
        [errIndL2, errIndH1weighted, errIndEdge] = hierarchical_p(u, uest, no_basis_z, no_basis_mu, elem, nodes, edges, edge2elem, int_edges, Jmp_fine, HF, phi, dphi, x, w, xc, wc, extra_z, extra_mu);
        err_ind = errIndL2 + errIndH1weighted;

    elseif uniform == 0 && strcmp(estimator, 'hierarchical_h')
        Cie = get_Cie(no_basis_z-1, phi, dphi, x, w); % compute constant from inverse inequality, cf (17)
        Cdt = 1+sqrt(Cie); % compute discrete trace constant, cf Lemma 2.
        alpha_penalty = (1/2+Cdt);
        % solve variational problem: for current mesh degree
        [u, x, w, xc, wc, phi, dphi, Jmp, IP, Penalty, int_edges, edge2elem, HF, sys] = solve_SIP(elem, nodes, edges, sigmas, sigmat, qx, gLx, gRx, no_basis_z, no_basis_mu, 2*alpha_penalty, output, edge2elem, int_edges);
        % refine mesh once
        elem_h = elem;
        nodes_h = nodes;
        edges_h = edges;
        edge2elem_h = edge2elem;
        int_edges_h = int_edges;
        for tbr = flip(1:size(elem, 2))
            [elem_h, nodes_h, edges_h, edge2elem_h, int_edges_h] = refine_quad(elem_h, nodes_h, edges_h, edge2elem_h, tbr);
        end
        % solve variational problem on refined mesh to compute error
        % estimator
        [uest, ~, ~, ~, ~, ~, ~, Jmp_fine, ~, Penalty_fine, ~, ~, ~, sys_fine] = solve_SIP(elem_h, nodes_h, edges_h, sigmas, sigmat, qx, gLx, gRx, no_basis_z+extra_z, no_basis_mu+extra_mu, alpha_penalty, output, edge2elem_h, int_edges_h);
        % compute error estimator
        [errIndL2, errIndH1weighted] = hierarchical_h(u, uest, no_basis_z, no_basis_mu, elem, phi, dphi, x, w, xc, wc, extra_z, extra_mu);
        err_ind = errIndL2 + errIndH1weighted;
    end


    %% Compute approximation errors

    % Compute interpolant of exact solution to grid (L2-projection)
    Ihu = getInterpolant(ux, elem, x, w, xc, wc, phi, no_basis_z, no_basis_mu);
    % Compute projection error and Galerkin approximation error in various norms
    [approx_err_L2, approx_err_L2_unsc, interp_err_L2, approx_err_derivative, approx_err_derivative_unsc, interp_err_derivative] = compute_errors(u, Ihu, ux, dux, elem, x, w, xc, wc, phi, dphi, no_basis_z, no_basis_mu);

    % Compute some error terms
    err_interp_Gal_Vh = Ihu(:)-u(:); % error between DG solution u and interpolant Ihu
    err_interp_Gal_Vh = sqrt(err_interp_Gal_Vh'*IP*err_interp_Gal_Vh); % Vh norm err_interp_Gal_Vh
    interp_error_Vh = sqrt(abs(sum(interp_err_derivative)+sum(interp_err_L2)+(Ihu(:)'*Penalty*Ihu(:)))); % Vh norm of interpolation error
    approx_error_Vh = sqrt(sum(approx_err_derivative+approx_err_L2))+abs(u(:)'*Penalty*u(:)); % Vh norm of approximation error

    %% Plotting
    if plotting
        % Plot mean of numerical solution
        figure(2)
        clf()
        patch('Faces', elem(1:4, :)', 'Vertices', nodes', 'FaceColor', 'flat', 'FaceVertexCData', u(1,:)', 'EdgeAlpha', 0.5, 'LineWidth', 1.25)
        title('Mean of $u_h$ per element', 'interpreter', 'Latex')
        colorbar

        % Plot broken H1-error between numerical and exact solution in log-10 scale
        figure(3)
        clf()
        patch('Faces', elem(1:4, :)', 'Vertices', nodes', 'FaceColor',' flat', 'FaceVertexCData', log10(sqrt(approx_err_L2_unsc+approx_err_derivative_unsc)), 'EdgeAlpha', 0.5, 'LineWidth', 1.25)
        % clim([-15 max(log10(sqrt(err_unsc)+1e-15))])
        clim([min(log10(sqrt(approx_err_L2_unsc+approx_err_derivative_unsc)+1e-15)) max(log10(sqrt(approx_err_L2_unsc+approx_err_derivative_unsc)+1e-15))])
        title('$\log_{10}$ of broken $H_1$-error per element', 'interpreter', 'Latex')
        % title('log10 of L2+H1-err per element')
        colorbar
        title_picture = 'Error indicator.png';
        print(title_picture, '-dpng','-r300');

        % Plot exact solution
        figure(4)
        clf()
        patch('Faces', elem(1:4, :)', 'Vertices', nodes', 'FaceColor', 'flat', 'faceVertexCData', Ihu(1, :).', 'EdgeAlpha', 0.5, 'LineWidth', 1.25)
        title('Midpoint value of exact solution per element')
        colorbar

        % Plot error indicator
        if uniform == 0
            figure(5)
            clf()
            patch('Faces', elem(1:4, :)', 'Vertices', nodes', 'FaceColor', 'flat', 'FaceVertexCData', log10(sqrt(err_ind)), 'EdgeAlpha', 0.5, 'LineWidth', 1.25)
            clim([min(log10(sqrt(err_ind)+1e-15)) max(log10(sqrt(err_ind)+1e-15))])
            title('$log_{10}$ of error indicator per element', 'interpreter', 'Latex')
            colorbar
        end
    end

    if uniform == 0
        % Doerfler marking
        [ind_s,i] = sort(err_ind, 'descend');
        ind = find((cumsum(ind_s)/sum(ind_s) > Theta), 1, 'first');
        if isempty(ind)
            ind = size(elem, 2);
        end
        refine = flip(i(1:ind));
        fprintf('elem = %5d, deg(z) = %d, deg(mu) = %d, |u-uh|_L2 = %1.2e, |u-I_h(u)|_L2 = %1.2e, |u-u_h|_Vh = %1.2e, |u_h-I_h(u)|_Vh = %1.2e, |u-I_h(u)|_Vh = %1.2e, |err_ind|_2 = %1.2e, ratio = %1.2d\n', size(elem, 2), no_basis_z-1, no_basis_mu-1, sqrt(sum(approx_err_L2)), sqrt(sum(interp_err_L2)), approx_error_Vh, err_interp_Gal_Vh, interp_error_Vh, sqrt(sum(err_ind)), sqrt(sum(err_ind))/approx_error_Vh)
    else
        refine = flip(1:size(elem, 2));
        fprintf('elem = %5d, deg(z) = %d, deg(mu) = %d, |u-uh|_L2 = %1.2e, |u-I_h(u)|_L2 = %1.2e, |u-u_h|_Vh = %1.2e, |u_h-I_h(u)|_Vh = %1.2e, |u-I_h(u)|_Vh = %1.2e\n', size(elem, 2), no_basis_z-1, no_basis_mu-1, sqrt(sum(approx_err_L2)), sqrt(sum(interp_err_L2)), approx_error_Vh, err_interp_Gal_Vh, interp_error_Vh)
    end

    GalL2 = [GalL2, sqrt(sum(approx_err_L2))];
    GalVh = [GalVh, approx_error_Vh];
    dofs = [dofs, size(elem,2)*no_basis];

end

f = @(x,p)((p+1)*p)^(p/2)*1./x.^(p/2); % optimal rate derived from h^p, x are the dofs.
p = no_basis_z-1;

if strcmp(estimator, 'hierarchical_p')
    if test_case == 6
        figure(10+p)
        loglog(dofs, GalVh, '*k', 'MarkerSize',12)
        hold on
        loglog(dofs, estimator_values,'ok','MarkerSize', 10)
        loglog(dofs, f(dofs,p)/f(dofs(1),p)*GalVh(1), '--k', 'LineWidth', 2)
        hold off
        lgd = legend('$\|u_{1}-u_{\mathcal{T}, k}\|_{\mathcal{T}}$', '$\|\zeta_p\|_{\mathcal{T}}$', 'optimal rate', 'interpreter', 'latex');
        lgd.FontSize = 16;
        lgd.Location = 'Best';
        set(gca, 'fontsize', 16), set(gcf, 'color', 'w')
        axis equal
        title(strcat(['k = ', num2str(k)]), 'interpreter', 'latex')
        title_picture = strcat(['u1, ', estimator, ', k = ', num2str(k), ' equal axis.png']);
        print(title_picture, '-dpng','-r300')

        figure(20+p)
        loglog(dofs, GalVh, '*k', 'MarkerSize',12)
        hold on
        loglog(dofs, estimator_values,'ok','MarkerSize', 10)
        loglog(dofs, f(dofs,p)/f(dofs(1),p)*GalVh(1), '--k', 'LineWidth', 2)
        hold off
        lgd = legend('$\|u_{1}-u_{\mathcal{T}, k}\|_{\mathcal{T}}$', '$\|\zeta_p\|_{\mathcal{T}}$', 'optimal rate', 'interpreter', 'latex');
        lgd.FontSize = 16;
        lgd.Location = 'Best';
        set(gca, 'fontsize', 16), set(gcf, 'color', 'w')
        title(strcat(['k = ', num2str(k)]), 'interpreter', 'latex')
        title_picture = strcat(['u1, ', estimator, ', k = ', num2str(k), '.png']);
        print(title_picture, '-dpng','-r300')

        filename = strcat(['u1, ', estimator, ', k = ', num2str(k), ', VARIABLES']);
        save(filename)

    elseif test_case == 3
        figure(10+p)
        loglog(dofs, GalVh, '*k', 'MarkerSize',12)
        hold on
        loglog(dofs, estimator_values,'ok','MarkerSize', 10)
        loglog(dofs, f(dofs,p)/f(dofs(1),p)*GalVh(1), '--k', 'LineWidth', 2)
        hold off
        lgd = legend('$\|u_{2}-u_{\mathcal{T}, k}\|_{\mathcal{T}}$', '$\|\zeta_p\|_{\mathcal{T}}$', 'optimal rate', 'interpreter', 'latex');
        lgd.FontSize = 16;
        lgd.Location = 'Best';
        set(gca, 'fontsize', 16), set(gcf, 'color', 'w')
        axis equal
        title(strcat(['k = ', num2str(k)]), 'interpreter', 'latex')
        title_picture = strcat(['u2, ', estimator, ', k = ', num2str(k), ' equal axis.png']);
        print(title_picture, '-dpng','-r300')

        figure(20+p)
        loglog(dofs, GalVh, '*k', 'MarkerSize',12)
        hold on
        loglog(dofs, estimator_values,'ok','MarkerSize', 10)
        loglog(dofs, f(dofs,p)/f(dofs(1),p)*GalVh(1), '--k', 'LineWidth', 2)
        hold off
        lgd = legend('$\|u_{2}-u_{\mathcal{T}, k}\|_{\mathcal{T}}$', '$\|\zeta_p\|_{\mathcal{T}}$', 'optimal rate', 'interpreter', 'latex');
        lgd.FontSize = 16;
        lgd.Location = 'Best';
        set(gca, 'fontsize', 16), set(gcf, 'color', 'w')
        title(strcat(['k = ', num2str(k)]), 'interpreter', 'latex')
        title_picture = strcat(['u2, ', estimator, ', k = ', num2str(k), '.png']);
        print(title_picture, '-dpng','-r300')

        filename = strcat(['u2, ', estimator, ', k = ', num2str(k), ', VARIABLES']);
        save(filename)

    end
elseif strcmp(estimator, 'hierarchical_h')
    if test_case == 6
        figure(10+p)
        loglog(dofs, GalVh, '*k', 'MarkerSize',12)
        hold on
        loglog(dofs, estimator_values,'ok','MarkerSize', 10)
        loglog(dofs, f(dofs,p)/f(dofs(1),p)*GalVh(1), '--k', 'LineWidth', 2)
        hold off
        lgd = legend('$\|u_{1}-u_{\mathcal{T}, k}\|_{\mathcal{T}}$', '$\|\zeta_h\|_{\mathcal{T}}$', 'optimal rate', 'interpreter', 'latex');
        lgd.FontSize = 16;
        lgd.Location = 'Best';
        set(gca, 'fontsize', 16), set(gcf, 'color', 'w')
        axis equal
        title(strcat(['k = ', num2str(k)]), 'interpreter', 'latex')
        title_picture = strcat(['u1, ', estimator, ', k = ', num2str(k), ' equal axis_new.png']);
        print(title_picture, '-dpng','-r300')

        figure(20+p)
        loglog(dofs, GalVh, '*k', 'MarkerSize',12)
        hold on
        loglog(dofs, estimator_values,'ok','MarkerSize', 10)
        loglog(dofs, f(dofs,p)/f(dofs(1),p)*GalVh(1), '--k', 'LineWidth', 2)
        hold off
        lgd = legend('$\|u_{1}-u_{\mathcal{T}, k}\|_{\mathcal{T}}$', '$\|\zeta_h\|_{\mathcal{T}}$', 'optimal rate', 'interpreter', 'latex');
        lgd.FontSize = 16;
        lgd.Location = 'Best';
        set(gca, 'fontsize', 16), set(gcf, 'color', 'w')
        title(strcat(['k = ', num2str(k)]), 'interpreter', 'latex')
        title_picture = strcat(['u1, ', estimator, ', k = ', num2str(k), '_new.png']);
        print(title_picture, '-dpng','-r300')

        % filename = strcat(['u1, ', estimator, ', k = ', num2str(k), ', VARIABLES']);
        % save(filename)

    elseif test_case == 3
        figure(10+p)
        loglog(dofs, GalVh, '*k', 'MarkerSize',12)
        hold on
        loglog(dofs, estimator_values,'ok','MarkerSize', 10)
        loglog(dofs, f(dofs,p)/f(dofs(1),p)*GalVh(1), '--k', 'LineWidth', 2)
        hold off
        lgd = legend('$\|u_{2}-u_{\mathcal{T}, k}\|_{\mathcal{T}}$', '$\|\zeta_h\|_{\mathcal{T}}$', 'optimal rate', 'interpreter', 'latex');
        lgd.FontSize = 16;
        lgd.Location = 'Best';
        set(gca, 'fontsize', 16), set(gcf, 'color', 'w')
        axis equal
        title(strcat(['k = ', num2str(k)]), 'interpreter', 'latex')
        title_picture = strcat(['u2, ', estimator, ', k = ', num2str(k), ' equal axis_new.png']);
        print(title_picture, '-dpng','-r300')

        figure(20+p)
        loglog(dofs, GalVh, '*k', 'MarkerSize',12)
        hold on
        loglog(dofs, estimator_values,'ok','MarkerSize', 10)
        loglog(dofs, f(dofs,p)/f(dofs(1),p)*GalVh(1), '--k', 'LineWidth', 2)
        hold off
        lgd = legend('$\|u_{2}-u_{\mathcal{T}, k}\|_{\mathcal{T}}$', '$\|\zeta_h\|_{\mathcal{T}}$', 'optimal rate', 'interpreter', 'latex');
        lgd.FontSize = 16;
        lgd.Location = 'Best';
        set(gca, 'fontsize', 16), set(gcf, 'color', 'w')
        title(strcat(['k = ', num2str(k)]), 'interpreter', 'latex')
        title_picture = strcat(['u2, ', estimator, ', k = ', num2str(k), ' _new.png']);
        print(title_picture, '-dpng','-r300')

        %filename = strcat(['u2, ', estimator, ', k = ', num2str(k), ', VARIABLES']);
        %save(filename)
    end
end