function [u,x,w,xc,wc,phi,dphi,Jmp,IP,Penalty,int_edges,edge2elem,HF,sys]=solve_SIP(elem,nodes,edges,sigmas,sigmat,q_source,gL_source,gR_source,no_basis_z,no_basis_mu,alpha_penalty,output,edge2elem,int_edges)
%% This code accompanies the manuscript
%% Riccardo Bardin, Fleurianne Bertrand, Olena Palii, Matthias Schlottbom
%% On a phase-space discontinuous Galerkin approximation for the radiative transfer equation in slab geometry.
%% Computational Methods in Applied Mathematics, doi: 10.1515/cmam-2023-0090
%% https://arxiv.org/abs/2201.06104

%% define basis functions by shifted Legendre (normalized on (0,1))
no_basis=no_basis_z*no_basis_mu;
% suppose we have a grid function u
% size(u)=[no_basis size(elem,2)], that is no_basis values per cell
[phi,dphi]=load_basis_functions();

zf=commonFinestPartition(elem); % used to apply the scattering operator

%% integration rules
[x,w]=lgwt(12,0,1); % Gauss-Legendre integration rule on (0,1) used of spatial integration.
zf_ip=diff(zf).*x(:)+zf(1:end-1); % composite Gauss-Legendre rule for common finest partition

[xc,wc]=lgwt(12,0,1/3); % composite Gauss-Legendre rule for mu integration
xc=[xc;xc+1/3;xc+2/3];
wc=[wc;wc;wc];

%% assemble system matrices
% assemble for example int mu^2 dz u dz v using for loops
row=zeros(no_basis^2,size(elem,2));
col=zeros(no_basis^2,size(elem,2));
val_A=zeros(no_basis^2,size(elem,2));
val_M=zeros(no_basis^2,size(elem,2));
val_flux=zeros(no_basis^2,size(elem,2));

%% loop over all elements to assemble local mass and stiffness matrix, etc
for it=1:size(elem,2)
    n1=elem(1,it); % first vertex of element
    n3=elem(3,it); % third vertex of element
    h=nodes(1,n3)-nodes(1,n1); % hz ( but h=hz=hmu here)
    mu_b=nodes(2,n1);
    z_l =nodes(1,n1);

    mu_ip = h*xc+ mu_b; % transform integration points from (0,1) to (mu_b,mu_t)
    z_ip  = h*x + z_l;  % transform integration points from (0,1) to (z_l,z_r)

    % local element matrices in mu:
    loc_mass_mu_squared =zeros(no_basis_mu,no_basis_mu);
    loc_mass_mu         =zeros(no_basis_mu,no_basis_mu);
    for i=1:no_basis_mu
        for j=1:no_basis_mu
            loc_mass_mu(i,j)        = h*(wc'*(phi{i}(xc).*phi{j}(xc)));
            loc_mass_mu_squared(i,j)= h*(wc'*(phi{i}(xc).*phi{j}(xc).*mu_ip.^2));
        end
    end
    % local element matrices in z:
    loc_stiff_z=zeros(no_basis_z,no_basis_z);
    loc_mass_z =zeros(no_basis_z,no_basis_z);
    loc_mass_z_sigma_t =zeros(no_basis_z,no_basis_z);
    loc_integrated_flux=zeros(no_basis_z,no_basis_z);
    stip=sigmat(z_ip);
    for i=1:no_basis_z
        for j=1:no_basis_z
            loc_stiff_z(i,j)        =(1/h) *(w'*(dphi{i}(x).*dphi{j}(x)./stip));
            loc_mass_z(i,j)         =    h *(w'*( phi{i}(x).* phi{j}(x)));
            loc_mass_z_sigma_t(i,j) =    h *(w'*( phi{i}(x).* phi{j}(x).*stip));
            loc_integrated_flux(i,j)=       (w'*( dphi{i}(x).*phi{j}(x)./stip));
        end
    end
    val_A(:,it)=reshape(kron(loc_stiff_z,loc_mass_mu_squared),no_basis^2,1);              % int int mu^2 (dz u) (dz v)/sigma_t dz dmu
    val_M(:,it)=reshape(kron(loc_mass_z_sigma_t,loc_mass_mu),no_basis^2,1);               % int int sigma_t u v dz dmu
    val_flux(:,it)=reshape(kron(loc_integrated_flux/h,eye(no_basis_mu)),no_basis^2,1);    % (dz u)/sigma_t

    % indices for local mass and local stiffness matrix
    ind=((it-1)*no_basis+1):it*no_basis;
    row(:,it)=reshape(repmat(ind,no_basis,1),no_basis^2,1);
    col(:,it)=reshape(repmat(ind,1,no_basis),no_basis^2,1);
end
A=sparse(row,col,val_A,no_basis*size(elem,2),no_basis*size(elem,2));
M=sparse(row,col,val_M,no_basis*size(elem,2),no_basis*size(elem,2));
flux=sparse(row,col,val_flux,no_basis*size(elem,2),no_basis*size(elem,2));     % flux 1/sigma_t dz u
                    
if isempty(edge2elem)
    [edge2elem,int_edges]=edge_to_elem(elem,nodes,edges);
end
no_int_edges=length(int_edges);

%% Assemble average and jump operators
% compute jump operator: elements to interior faces
% compute average operator: elements to interior faces
% L^2 inner product on interior faces; Here, we have polynomials of degree
% no_basis_mu (ONB up to hF)
row_jmp=zeros(no_basis_mu,no_basis_z,no_basis_mu,2*no_int_edges);
col_jmp=zeros(no_basis_mu,no_basis_z,no_basis_mu,2*no_int_edges);
val_jmp=zeros(no_basis_mu,no_basis_z,no_basis_mu,2*no_int_edges);
row_avg=zeros(no_basis_mu,no_basis_z,no_basis_mu,2*no_int_edges);
col_avg=zeros(no_basis_mu,no_basis_z,no_basis_mu,2*no_int_edges);
val_avg=zeros(no_basis_mu,no_basis_z,no_basis_mu,2*no_int_edges);

row_faces  = zeros(no_basis_mu,no_basis_mu,no_int_edges);
col_faces  = zeros(no_basis_mu,no_basis_mu,no_int_edges);
val_l2f    = zeros(no_basis_mu,no_basis_mu,no_int_edges);
val_l2fm2  = zeros(no_basis_mu,no_basis_mu,no_int_edges);
val_l2f_unw= zeros(no_basis_mu,no_basis_mu,no_int_edges);
HF=zeros(no_int_edges,1);
for it=1:no_int_edges % loop over interior faces
    el_l=edge2elem(1,int_edges(it)); % left element
    el_r=edge2elem(2,int_edges(it)); % right element
    level_el_l=elem(5,el_l);
    h_el_l=1/2^level_el_l;
    level_el_r=elem(5,el_r);
    h_el_r=1/2^level_el_r;
    st_l = 1; %sigma_t left %TODO
    st_r = 1; %sigma_t right %TODO
    HF(it)=st_l*st_r*h_el_l*h_el_r/(st_l*h_el_l+st_r*h_el_r); %=1/(1/(st_l*h_el_l)+1/(st_r*h_el_r));

    n1=edges(1,int_edges(it)); n2=edges(2,int_edges(it));
    hF=norm(nodes(:,n1)-nodes(:,n2)); % face length
    
    % average {u} = 0.5*(u(z_{right}^{K_{left})+u(z_{left}^{K_{right}))
    %             = 0.5*(sum_{j=1}^no_basis_mu
    %             (sum_{i=1}^no_basis_z u^l_{i,j}*phi{i}(1) phi^l{j}(mu) + u^r_{i,j}*phi^r{j}(mu)*phi{i}(0)))
    % --> note that on the left and right element the basis functions
    % evaluate differently in mu for hanging nodes. Compute the values of the mu-basis
    % functions in the mu integration points on the face

    % jump: [u] = u(z_{right}^{K_left}) - u(z_{left}^{K_right})
    %           = (sum_{j=1}^no_basis_mu
    %             (sum_{i=1}^no_basis_z u^l_{i,j}*phi{i}(1)phi^l{j}(mu) - u^r_{i,j}*phi{j}(mu)^r*phi{i}(0)))
    
    ind_l = reshape((el_l-1)*no_basis+1:el_l*no_basis,no_basis_mu,no_basis_z); %indices from left element
    ind_r = reshape((el_r-1)*no_basis+1:el_r*no_basis,no_basis_mu,no_basis_z); %indices from right element

    n1_l   = elem(1,el_l);  n1_r   = elem(1,el_r); % first node of left/right element
    mu_b_l = nodes(2,n1_l); mu_b_r = nodes(2,n1_r); % mu-component of first node
    
    mu_b=min(nodes(2,[n1 n2]));
    mu_ip = hF*xc+mu_b; %transform integration points from (0,1) to (mu_b,mu_t)
    
    % the shorter edge is a subinterval of the corresponding faces of the left and right element.
    % map integration points on interior face back to subinterval of (0,1)
    mu_ip_left = (mu_ip-mu_b_l)/h_el_l;  % relative to face of left element
    mu_ip_right= (mu_ip-mu_b_r)/h_el_r; % relative to face of right element

    % interpolation operator taking restrictions of u_h^left and u_h^right to polynomial on face
    trace_left_local=zeros(no_basis_mu,no_basis_mu);
    trace_right_local=zeros(no_basis_mu,no_basis_mu);
    for i=1:no_basis_mu
        for j=1:no_basis_mu
            trace_left_local(i,j) = wc'*(phi{i}(xc).*phi{j}(mu_ip_left)); % left
            trace_right_local(i,j)= wc'*(phi{i}(xc).*phi{j}(mu_ip_right)); % right
        end
    end
    %
    for i=1:no_basis_mu
        for j=1:no_basis_z
            for k=1:no_basis_mu
                % avg and jmp contribution from right element
                val_avg(i,j,k,it)= 0.5*phi{j}(0) * trace_right_local(i,k);
                val_jmp(i,j,k,it)=    -phi{j}(0) * trace_right_local(i,k);
                row_avg(i,j,k,it)= no_basis_mu*(it-1) + i;
                row_jmp(i,j,k,it)= no_basis_mu*(it-1) + i;
                col_avg(i,j,k,it)= ind_r(k,j);
                col_jmp(i,j,k,it)= ind_r(k,j);
                % avg and jmp contribution from left element
                val_avg(i,j,k,it+no_int_edges)= 0.5*phi{j}(1) * trace_left_local(i,k);
                val_jmp(i,j,k,it+no_int_edges)=     phi{j}(1) * trace_left_local(i,k);
                row_avg(i,j,k,it+no_int_edges)= no_basis_mu*(it-1) + i;
                row_jmp(i,j,k,it+no_int_edges)= no_basis_mu*(it-1) + i;
                col_avg(i,j,k,it+no_int_edges)= ind_l(k,j);
                col_jmp(i,j,k,it+no_int_edges)= ind_l(k,j);
            end
        end
    end
    % integrate mass matrices for the interior faces
    for i=1:no_basis_mu
        for j=1:no_basis_mu
            val_l2f_unw(i,j,it) = hF * wc'*(phi{i}(xc).*phi{j}(xc)); % (unweighted) L2-inner product on faces: int_F u v dF
            val_l2f(i,j,it)     = hF * wc'*(phi{i}(xc).*phi{j}(xc).*mu_ip); % mu-weighted L2-inner product on faces: int_F mu u v dF
            val_l2fm2(i,j,it)   = hF * wc'*(phi{i}(xc).*phi{j}(xc).*mu_ip.^2); % mu^2-weighted L2-inner product on faces: int_F mu^2 u v dF
            row_faces(i,j,it) = no_basis_mu*(it-1) + i;
            col_faces(i,j,it) = no_basis_mu*(it-1) + j;
        end
    end
end
Avg=sparse(row_avg(:),col_avg(:),val_avg(:),no_basis_mu*no_int_edges,no_basis*size(elem,2)); % Avg*u gives {u} on interior faces
Jmp=sparse(row_jmp(:),col_jmp(:),val_jmp(:),no_basis_mu*no_int_edges,no_basis*size(elem,2)); % Jmp*u gives [u] on interior faces
L2F     =sparse(row_faces(:),col_faces(:),val_l2f(:)    ,no_basis_mu*no_int_edges,no_basis_mu*no_int_edges); % L2(mu) integrals over interior faces
L2Fm2   =sparse(row_faces(:),col_faces(:),val_l2fm2(:)  ,no_basis_mu*no_int_edges,no_basis_mu*no_int_edges); % L2(mu^2) integrals over interior faces

%% assemble penalty term
HF2=reshape(repmat(HF',no_basis_mu,1),no_basis_mu*no_int_edges,1);
Penalty=Jmp'*L2F*(diag(sparse(1./HF2))*Jmp);

%% assemble boundary bilinear form
bnd_edges=setdiff(1:size(edges,2),int_edges);
row=zeros(no_basis^2,length(bnd_edges));
col=row;
val=row;
for it=1:length(bnd_edges)
    idx=bnd_edges(it);
    el_l=edge2elem(1,idx);
    el_r=edge2elem(2,idx);

    n1=edges(1,idx); n2=edges(2,idx);
    hF=norm(nodes(:,n1)-nodes(:,n2)); % face length
    mu_b=min(nodes(2,[n1 n2]));
    mu_ip = hF*xc+mu_b; %transform integration points from (0,1) to (mu_b,mu_t)

    if el_l<0 % left_boundary
        nei=el_r; %neighbour element
        zb=0;
    else % right boundary
        nei=el_l; %neighbour element
        zb=1;
    end

    loc_mu_int=zeros(no_basis_mu,no_basis_mu);
    loc_z_eval=zeros(no_basis_z,no_basis_z);
    for i=1:no_basis_mu
        for j=1:no_basis_mu
            loc_mu_int(i,j)=hF * wc'*(phi{i}(xc).*phi{j}(xc).*mu_ip);
        end
    end
    for i=1:no_basis_z
        for j=1:no_basis_z
            loc_z_eval(i,j)=phi{i}(zb)*phi{j}(zb);
        end
    end
    val(:,it)=reshape(kron(loc_z_eval,loc_mu_int),no_basis^2,1);

    % indices for local mass and local stiffness matrix
    ind=(nei-1)*no_basis+1:nei*no_basis;
    row(:,it)=reshape(repmat(ind,no_basis,1),no_basis^2,1);
    col(:,it)=reshape(repmat(ind,1,no_basis),no_basis^2,1);
end
B=sparse(row,col,val,no_basis*size(elem,2),no_basis*size(elem,2));

%% overall bilinear form without scattering
XX=(Avg*flux)'*L2Fm2*Jmp;
sys=A+B+M-XX-XX'+alpha_penalty*Penalty;

%% assemble interior source term and interpolant of exact solution to grid (L2-projection)
q=zeros(no_basis,size(elem,2));
% compute interpolation of q and u_exact into broken FEM space
for it=1:size(elem,2)
    l=elem(5,it);
    h=1/2^l;
    [i,j]=ind2sub([2^l,2^l],elem(6,it));

    z_l=(i-1)*h;  % z_r= i*h;
    mu_b=(j-1)*h; % mu_t= j*h;

    x_m=h*x+z_l;        % map integration points to I_z
    mu_ip=h*xc+mu_b;    % map integration points to I_mu

    q_loc=zeros(no_basis_mu,no_basis_z);
    for i=1:no_basis_z
        for j=1:no_basis_mu
            q_loc(j,i)= h^2 *(phi{i}(x).*w)'*q_source(x_m,mu_ip')*(wc.*phi{j}(xc));
        end
    end
    q(:,it)=q_loc(:);
end
q=q(:);

%% assemble boundary source term
g=zeros(no_basis,size(elem,2));
for it=1:length(bnd_edges)
    el_l=edge2elem(1,bnd_edges(it));
    el_r=edge2elem(2,bnd_edges(it));
    n1=edges(1,bnd_edges(it)); n2=edges(2,bnd_edges(it)); % node indices describing face
    hF=norm(nodes(:,n1)-nodes(:,n2)); % face length

    if el_l<0 % left_boundary
        n1_elem=elem(1,el_r);
        mu_b=nodes(2,n1_elem);
        mu_ip=hF*xc+mu_b; % map integration points to I_mu

        g_loc=zeros(no_basis_mu,no_basis_z);
        for i=1:no_basis_z
            for j=1:no_basis_mu
                g_loc(j,i)=phi{i}(0) * hF * wc'*(mu_ip.*gL_source(mu_ip).*phi{j}(xc));
            end
        end
        g(:,el_r)=g_loc(:);
    else % right boundary
        n1_elem=elem(1,el_l);
        mu_b=nodes(2,n1_elem);
        mu_ip=hF*xc+mu_b; % map integration points to I_mu
        g_loc=zeros(no_basis_mu,no_basis_z);

        for i=1:no_basis_z
            for j=1:no_basis_mu
                g_loc(j,i)=phi{i}(1) * hF * wc'*(mu_ip.*gR_source(mu_ip).*phi{j}(xc));
            end
        end
        g(:,el_l)=g_loc(:);
    end
end
g=g(:);

%% create mesh structure that carries the common finest partition.
mesh.nodes=nodes;
mesh.elem=elem;
mesh.phi=phi;
mesh.no_basis_z=no_basis_z;
mesh.no_basis_mu=no_basis_mu;
mesh.x=x;
mesh.w=w;
mesh.zf=zf;
mesh.zf_ip=zf_ip;

K=@(u) apply_scattering(mesh,sigmas,u); % apply scattering operator

%% compute continuity and coercivity bounds numerically
IP=A+B+M+Penalty; IP=0.5*(IP+IP');
if size(elem,2)<10000 && output
    sys2=(0.5)*(sys+sys');
    dim=min(size(IP,2),25);
    [~,dl]=eigs(sys2,IP,1,'lm','IsSymmetricDefinite',1,'SubspaceDimension',dim,'MaxIterations',10000);
    [~,dm]=eigs(sys2,IP,1,'sm','IsSymmetricDefinite',1,'SubspaceDimension',dim,'MaxIterations',10000);
    if output
        fprintf('continuity bound %f; coercivity bound %f\n',dl,dm);
    end
end

%% Solve the linear system via source iteration
u=0*q; % initialization
if output
    fprintf('Enter solve: Source iteration ')
end
for it=1:1000
    u0=u;
    u=sys\(K(u)+q+g);
    res=u-u0;
    if sqrt(res'*M*res)<1e-10
        break
    end
end
if output
    fprintf(' done in %3d iterations with res=%1.2e\n',it,sqrt(res'*M*res))
end
u=reshape(u,no_basis,size(elem,2));